/*
 
 em.inc
 2010 Samuel Benge
 
 This is an edge map (proximity pattern) for a pigment. It can be used to quickly add
 proximity effects to height_fields. It returns a pigment_pattern, so it can be used
 anywhere. The pattern is 2D and will face the +z direction. Also includes a pigment
 blur macro.
 
 
 The Macros
 
 em(pgmt, amnt, amtr, qual)
  * pgmt = a pigment
  * amnt = edge radius
  * amtr = jitter
  * qual = quality (0 to 1)
 
 This will make an edge map from any pigment.
 
 
 em_pigment_blur(pgmt, amnt, amtr, qual)
  * pgmt = a pigment
  * amnt = blur amount
  * amtr = jitter
  * qual = quality (0 to 1)
 
 This will blur a pigment in three dimensions. It can be used as a proximity pattern if
 you supply it with a white/black object pigment.
 
 
 Example code:
 
  #declare my_image =
  pigment{
   image_map{
    png "../some_image.png"
    interpolate 2
   }
  }
 
  pigment{
   em(my_image, 0.01, 0.005, 0.3)
  }
 
*/

// pigment blur
#macro em_pigment_blur(pgmt, amnt, amtr, qual)
 #if(qual<0) #local qual=0; #end
 #if(qual>1) #local qual=1; #end
 // an array containing a 3D spherical grid of points
 #local pts_sphere=array[251];
 #local ct = 0;
 #local extent = 3;
 #local Z=-extent;
 #while(Z<=extent)
  #local Y=-extent;
  #while(Y<=extent)
   #local X=-extent;
   #while(X<=extent)
    #if(sqrt(X*X+Y*Y+Z*Z)<=extent+1)
     #local pts_sphere[ct]=<X,Y,Z>*1/(extent+1);
     #local ct=ct+1;
    #end
    #local X=X+1;
   #end
   #local Y=Y+1;
  #end
  #local Z=Z+1;
 #end
 
 // the blurred pigment
 pigment_pattern{
  average
  pigment_map{
   #local R = seed(1002);
   #local ct = 0;
   #while(ct<dimension_size(pts_sphere,1))
    [1
     pgmt
     translate pts_sphere[ct]*amnt-<rand(R)-rand(R),rand(R)-rand(R),rand(R)-rand(R)>*amtr
    ]
    #local ct=ct+1/qual;
   #end
  }
 }
#end

// edge map pigment
#macro em(pgmt, amnt, amtr, qual)
 #local f_plane = function{select(z,0,1)}
 #local f_image = function{pigment{pgmt}}
 #local f_blurred_height =
 function{
  pigment{
   em_pigment_blur(
    pigment{
     function{
      f_plane(x,y,z-f_image(x,y,0).gray)
     }
    }
    ,amnt,amtr,qual
   )
  }
 }
 pigment_pattern{
  function{ f_blurred_height(x,y,z+f_image(x,y,0).gray).gray }
  warp{planar}
 }
#end